#!/usr/bin/python3

"""
Name: GSettings-Accounts - [The GSettings Project](https://gsettings.tuxfamily.org)
Brief: user password change and accounts management
Author: Frank ENDRES <frankendres.tuxfamily.org> (disabled <first@last.fr>
Version: v0.4
License: CeCILL - http://www.cecill.info/licences/Licence_CeCILL_V2.1-en.html (GPL compatible)
"""

from typing import Dict, List, Tuple
import sys #for those who can't guess, this is system programming	;-)
import os
import stat
import shutil
import subprocess
import pwd
import grp #as I said, this is system programming 					;-)
import re

import gettext; _ = gettext.gettext
import gi
gi.require_version('Gtk', '3.0')
# ~ gi.require_version('Gtk', '4.0')
from gi.repository import GLib, Gtk, Gio
from gi.repository import Gtk, Gio, GdkPixbuf
import warnings; warnings.filterwarnings("ignore", category=DeprecationWarning) #Gtk.TreeView, Gtk.ComboBoxText, Gtk.FileChooserDialog, Gtk.ComboBoxText





# GTK3 / 4 compatibility layer -> code is written in GTK4 style
# to toogle with Geany: select the whole Compat bloc, then <Ctrl> + <e>
class CompatApplicationWindow(Gtk.ApplicationWindow): #GTK3
	def set_child(self, child): #renamed method
		for content in self.get_children(): #keep header bar and remove other content
			if not isinstance(content, CompatHeaderBar):
				self.remove(content)
		self.add(child)

	def present(self): #renamed method
		self.show_all()
		super().present()
# ~ class CompatApplicationWindow(Gtk.ApplicationWindow):
	# ~ pass #GTK4


class CompatWindow(Gtk.Window): #GTK3
	def set_child(self, child): #renamed method
		self.add(child)

	def present(self): #renamed method
		self.show_all()
		super().present()

	def set_default_widget(self, widget: Gtk.Widget):
		widget.set_can_default(True)
		self.set_default(widget)
# ~ class CompatWindow(Gtk.Window): #GTK4
	# ~ pass


class CompatScrolledWindow(Gtk.ScrolledWindow): #GTK3
	def set_child(self, child): #renamed method
		self.add(child)
# ~ class CompatScrolledWindow(Gtk.ScrolledWindow): #GTK4
	# ~ pass


class CompatBox(Gtk.Box): #GTK3
	def append(self, widget): #renamed method
		self.pack_start(widget, False, False, 0)
# ~ class CompatBox(Gtk.Box): #GTK4
	# ~ pass


class CompatButton(Gtk.Button): #GTK3
	def new_from_icon_name(icon_name): #changed method profile
		return Gtk.Button.new_from_icon_name(icon_name, Gtk.IconSize.BUTTON)
# ~ class CompatButton(Gtk.Button): #GTK4
	# ~ def new_from_icon_name(icon_name):
		# ~ return Gtk.Button.new_from_icon_name(icon_name)

class CompatAboutDialog(Gtk.AboutDialog): #GTK3
	def set_visible(self, visible): #renamed method
		if visible == True:
			super().run()
			self.destroy()
# ~ class CompatAboutDialog(Gtk.AboutDialog): #GTK4
	# ~ pass


class CompatPasswordEntry(Gtk.Box): #GTK3
	def __init__(self):
		super().__init__(orientation = Gtk.Orientation.HORIZONTAL, spacing=10)
		self.password_entry = Gtk.Entry()
		self.password_entry.set_visibility(False)
		self.password_entry.set_activates_default(True)
		self.pack_start(self.password_entry, True, True, 0)
		show_password_button = CompatButton.new_from_icon_name('media-optical-symbolic')
		show_password_button.set_tooltip_text(_('show / hide password'))
		self.pack_start(show_password_button, False, False, 0)
		show_password_button.connect('clicked', lambda bt: self.password_entry.set_visibility(not self.password_entry.get_visibility()))

	def get_text(self) -> str:
		return self.password_entry.get_text()

	def connect(self, event, callback):
		self.password_entry.connect(event, callback)

# ~ class CompatPasswordEntry(Gtk.PasswordEntry): #GTK4
	# ~ def __init__(self):
		# ~ super().__init__()
		# ~ self.set_show_peek_icon(True)


class CompatHeaderBar(Gtk.HeaderBar): #GTK3
	def __init__(self):
		super().__init__()
		self.set_show_close_button(True)

	def set_title_widget(self, label):
		self.props.title = label.get_label()
# ~ class CompatHeaderBar(Gtk.HeaderBar): #GTK4
	# ~ pass





MEMBERSHIP_STORE_LOGIN = 0
MEMBERSHIP_STORE_TOGGLE = 1

class EditUser(Gtk.Grid):
	#CHOICE: uids and gids cannot be changed
	#CHOICE: home directory can't be changed (/home/login)

	def __init__(self, app: Gtk.Application, login: str, group: str):
		super().__init__()
		self.set_margin_start(10); self.set_margin_end(10); self.set_margin_top(10); self.set_margin_bottom(10)
		self.set_row_spacing(10)
		self.set_column_spacing(10)
		self.app = app
		self.infos = None
		if login != None: #existing user
			self.infos = get_user(login)
			self.add_widgets()
			self.reload(None)
		else: #new user
			self.infos = get_emty_user()
			self.add_widgets()
			if group != None:
				self.group_cb.set_active(self.user_groups.index(group))


	def add_widgets(self):
		login = get_username()
		# Name
		name_label = Gtk.Label(label=_('Name:'))
		self.attach(name_label, 0, 0, 1, 1)
		self.name_entry = Gtk.Entry()
		self.attach(self.name_entry, 1, 0, 1, 1)
		if not self.infos[LOGGED]:
			self.name_entry.connect('changed', self.on_name_entry_clicked)
		if login != 'root' and get_login_defs()['CHFN_RESTRICT'].find('f') == -1:
			self.name_entry.set_sensitive(False)

		# Login
		login_label = Gtk.Label(label=_('Username:'))
		self.attach(login_label, 0, 1, 1, 1)
		self.login_entry = Gtk.Entry()
		self.login_entry.connect('changed', lambda e: self.error_label.set_text(''))
		self.attach(self.login_entry, 1, 1, 1, 1)
		if self.infos[LOGGED] or login != 'root':
			self.login_entry.set_sensitive(False)

		# Password
		password_label = Gtk.Label(label=_('Password:'))
		self.attach(password_label, 0, 2, 1, 1)
		self.password_entry = CompatPasswordEntry()
		self.attach(self.password_entry, 1, 2, 1, 1)

		# Shell
		shell_label = Gtk.Label(label=_('Shell:'))
		self.attach(shell_label, 0, 3, 1, 1)
		self.shell_cb = Gtk.ComboBoxText() #FIXME: deprecated -> Gtk.DropDown
		self.shell_cb.set_entry_text_column(0)
		self.shells = load_shells()
		for shell in self.shells:
			self.shell_cb.append_text(shell)
		self.shell_cb.set_active(0)
		self.attach(self.shell_cb, 1, 3, 1, 1)

		# Group
		group_label = Gtk.Label(label=_('Group:'))
		self.attach(group_label, 2, 0, 1, 1)
		self.group_cb = Gtk.ComboBoxText()
		self.group_cb.set_entry_text_column(0)
		self.user_groups = [''] + load_user_groups()
		for group in self.user_groups:
			self.group_cb.append_text(group)
		self.group_cb.set_active(0)
		self.attach(self.group_cb, 3, 0, 1, 1)
		if login != 'root':
			self.group_cb.set_sensitive(False)

		# Membership
		self.membership_store = Gtk.ListStore(str, bool)
		self.membership_groups = load_membership_groups()
		for group in self.membership_groups:
			self.membership_store.append([group, False])
		membership_treeview = Gtk.TreeView(model=self.membership_store) #FIXME: deprecated -> Gtk.ListView
		membership_label = Gtk.TreeViewColumn(_('Membership'), Gtk.CellRendererText(), text=MEMBERSHIP_STORE_LOGIN)
		membership_treeview.append_column(membership_label)
		membership_toggle_renderer = Gtk.CellRendererToggle()
		membership_toggle_renderer.connect("toggled", self.on_membership_toggled)
		membership_toggle = Gtk.TreeViewColumn("Activer", membership_toggle_renderer, active=MEMBERSHIP_STORE_TOGGLE)
		membership_treeview.append_column(membership_toggle)
		membership_scroll = CompatScrolledWindow()
		membership_scroll.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
		membership_scroll.set_child(membership_treeview)
		self.attach(membership_scroll, 2, 1, 2, 4)
		if login != 'root':
			membership_treeview.set_sensitive(False)

		# Avatar:
		box = CompatBox(orientation = Gtk.Orientation.HORIZONTAL, spacing=10)
		# ~ box.set_halign(Gtk.Align.FILL) #FIXME: looks like in has no effect
		avatar_choose_button = Gtk.Button(label=_('Choose avatar'))
		avatar_choose_button.set_hexpand(True)
		avatar_choose_button.set_label(label=_('Choose avatar'))
		self.init_avatar_chooser()
		avatar_choose_button.connect("clicked", lambda bt: self.avatar_chooser.set_visible(True))
		box.append(avatar_choose_button)
		avatar_reset_button = CompatButton.new_from_icon_name('user-trash-symbolic')
		avatar_reset_button.set_tooltip_text(_('delete avatar'))
		avatar_reset_button.connect("clicked", self.on_avatar_unset)
		box.append(avatar_reset_button)
		self.attach(box, 4, 0, 2, 1)
		self.avatar_preview = Gtk.Image()
		self.avatar_preview.set_pixel_size(120)
		self.attach(self.avatar_preview, 4, 1, 2, 3)
		self.avatar = None

		# Message:
		self.error_label = Gtk.Label()
		self.attach(self.error_label, 0, 5, 6, 1)

		# Buttons
		cancel_button = Gtk.Button(label=_('Cancel'))
		cancel_button.connect('clicked', lambda button: self.app.on_end_edit_user())
		self.attach(cancel_button, 0, 6, 1, 1)
		if self.infos[LOGIN] == None:
			save_button = Gtk.Button(label=_('Create'))
			save_button.connect('clicked', self.on_save_clicked)
			self.attach(save_button, 5, 6, 1, 1)
		else:
			reload_button = Gtk.Button(label=_('Reload'))
			reload_button.connect('clicked', self.reload)
			self.attach(reload_button, 4, 6, 1, 1)
			save_button = Gtk.Button(label=_('Update'))
			if login == 'root':
				save_button.connect('clicked', self.on_save_clicked)
			else:
				save_button.connect('clicked', self.on_self_save_clicked)
			self.attach(save_button, 5, 6, 1, 1)


	def on_name_entry_clicked(self, entry: Gtk.Entry):
		auto_login = self.name_entry.get_text().lower().replace(' ', '')
		login = self.login_entry.get_text()
		cp_len = len(auto_login) - 1
		if auto_login[:cp_len] == login[:cp_len]:
			self.login_entry.set_text(auto_login)


	def on_membership_toggled(self, renderer, path: str):
		iter = self.membership_store.get_iter_from_string(path)
		self.membership_store[iter][MEMBERSHIP_STORE_TOGGLE] = not self.membership_store[iter][MEMBERSHIP_STORE_TOGGLE]


	def init_avatar_chooser(self):
		self.avatar_chooser = Gtk.FileChooserDialog(title= _('Choose avatar'), action=Gtk.FileChooserAction.OPEN) #FIXME: deprecated -> Gtk.FileDialog
		self.avatar_chooser.set_transient_for(self.app.win)
		image_filter = Gtk.FileFilter()
		image_filter.add_mime_type('image/*')
		self.avatar_chooser.set_filter(image_filter)
		self.avatar_chooser.add_buttons(_('Cancel'), Gtk.ResponseType.CANCEL, _('OK'), Gtk.ResponseType.ACCEPT)
		self.avatar_chooser.connect('response', self.on_avatar_dialog_response)


	def on_avatar_dialog_response(self, dialog: Gtk.FileChooser, resp: Gtk.ResponseType):
		if resp == Gtk.ResponseType.ACCEPT:
			self.avatar = dialog.get_file().get_path()
			self.display_avatar(self.avatar)
		dialog.set_visible(False)


	def on_avatar_unset(self, button: Gtk.Button):
		self.avatar = None
		self.avatar_preview.clear()

	def display_avatar(self, filename: str):
		pixbuf = GdkPixbuf.Pixbuf.new_from_file(filename) #FIXME: deprecated -> Gtk.Paintable
		scaled = pixbuf.scale_simple(120, 120, GdkPixbuf.InterpType.BILINEAR)
		self.avatar_preview.set_from_pixbuf(scaled)


	def reload(self, button: Gtk.Button):
		#CHOICE: don't call get_user() since the GUI could alse have to be rebuild
		self.name_entry.set_text(self.infos[COMMENT])
		self.name_entry.set_position(-1)
		self.login_entry.set_text(self.infos[LOGIN])
		self.login_entry.set_position(-1)
		if self.infos[SHELL] in self.shells:
			self.shell_cb.set_active(self.shells.index(self.infos[SHELL]))
		self.group_cb.set_active(self.user_groups.index(self.infos[GROUP]))
		membership = self.infos[MEMBERSHIP].split(',')
		for row in self.membership_store:
			if row[MEMBERSHIP_STORE_LOGIN] in membership:
				row[MEMBERSHIP_STORE_TOGGLE] = True
		if self.infos[AVATAR] != None:
			self.avatar = self.infos[AVATAR]
			self.display_avatar(self.infos[AVATAR])


	def on_save_clicked(self, button):
		group = self.user_groups[self.group_cb.get_active()]
		membership = []
		for row in self.membership_store:
			if row[MEMBERSHIP_STORE_TOGGLE] and row[MEMBERSHIP_STORE_LOGIN] != group:
				membership.append(row[MEMBERSHIP_STORE_LOGIN])
		infos = {
			LOGIN: self.login_entry.get_text(),
			UID: None if self.infos == None else self.infos[UID],
			GROUP: group,
			COMMENT: self.name_entry.get_text(),
			SHELL: self.shells[self.shell_cb.get_active()],
			PASSWORD: self.password_entry.get_text(),
			MEMBERSHIP: ','.join(membership),
			AVATAR: self.avatar,
		}
		if self.infos[LOGIN] == infos[LOGIN] or self.check_login(infos[LOGIN]):
			save_user(infos)
			self.app.on_end_edit_user()


	def on_self_save_clicked(self, button):
		self.password_dialog = CompatWindow(title=_('Authentication'))
		self.password_dialog.set_transient_for(self.app.win)
		box = CompatBox(orientation = Gtk.Orientation.VERTICAL, spacing=10)
		box.set_margin_start(10); box.set_margin_end(10); box.set_margin_top(10); box.set_margin_bottom(10)
		self.password_dialog.set_child(box)
		box.append(Gtk.Label(label=_('Please authenticate yourself to apply changes:')))
		self.authentication_entry = CompatPasswordEntry()
		self.authentication_entry.connect('changed', lambda e: self.password_error_label.set_text(''))
		box.append(self.authentication_entry)
		self.password_error_label = Gtk.Label(label = ' ')
		box.append(self.password_error_label)
		buttons = CompatBox(orientation = Gtk.Orientation.HORIZONTAL, spacing=10)
		cancel_button = Gtk.Button(label=_('Cancel'))
		cancel_button.set_hexpand(True)
		cancel_button.connect('clicked', lambda bt: self.password_dialog.set_visible(False))
		buttons.append(cancel_button)
		ok_button = Gtk.Button(label=_('OK'))
		ok_button.set_hexpand(True)
		ok_button.connect('clicked', self.on_password_dialog_auth)
		buttons.append(ok_button)
		self.password_dialog.set_default_widget(ok_button)
		box.append(buttons)

		comment = self.name_entry.get_text()
		shell = self.shells[self.shell_cb.get_active()]
		password = self.password_entry.get_text()
		self.auth_tries = 0
		if comment != self.infos[COMMENT] or shell != self.infos[SHELL] or password != '' or self.infos[AVATAR] != self.avatar:
			self.password_dialog.present()


	def on_password_dialog_auth(self, bt: Gtk.Button):
		comment = self.name_entry.get_text()
		shell = self.shells[self.shell_cb.get_active()]
		password = self.password_entry.get_text()
		authentication = self.authentication_entry.get_text()
		if check_password(self.infos[LOGIN], authentication): #TODO: run in a thread
			if comment != self.infos[COMMENT]:
				chfn_user(self.infos[LOGIN], authentication, comment)
			if shell != self.infos[SHELL]:
				chsh_user(self.infos[LOGIN], authentication, shell)
			if password != '':
				change_password(self.infos[LOGIN], authentication, password) #TODO: password change may be refused if too simple
			if self.avatar != get_avatar(self.infos[LOGIN]):
				set_avatar(self.infos[LOGIN], self.avatar)
			self.app.on_end_edit_user()
		else:
			self.auth_tries += 1
			if self.auth_tries == 3:
				self.app.on_end_edit_user()
			else:
				self.password_error_label.set_markup('<span foreground="red" weight="bold">' + _(f"Wrong password: {self.auth_tries}/3 failures.") + '</span>')


	def check_login(self, login: str) -> bool:
		status = check_name(login, False)
		message = [ None, _('Invalid login:') + _('only 32  _ , a-z, A-Z and 0-9 characters are allowed.'), _('Login is already assigned to a user.') ][status]
		if message != None:
			self.error_label.set_markup('<span foreground="red" weight="bold">' + message + '</span>')
		return True if message == None else False





NO_GROUP = '.'

class ManageLists(Gtk.Grid):
	def __init__(self, app: Gtk.Application, group: str):
		super().__init__(row_spacing=10, column_spacing=10)
		self.set_margin_start(10); self.set_margin_end(10); self.set_margin_top(10); self.set_margin_bottom(10)
		self.app = app
		self.add_widgets()
		self.select_group(group)


	def add_widgets(self):
		self.error_label = Gtk.Label()
		self.attach(self.error_label, 0, 2, 4, 1)

		# Groups
		self.groups_store = Gtk.ListStore(str)
		self.groups_store.append([NO_GROUP])
		for group in load_user_groups():
			self.groups_store.append([group])
		self.groups_treeview = Gtk.TreeView(model=self.groups_store)
		self.groups_treeview.set_activate_on_single_click(True)
		self.groups_treeview.connect('row-activated', self.on_group_selected)
		group_renderer = Gtk.CellRendererText(editable=True)
		group_renderer.connect('edited', self.on_group_edited)
		group_renderer.connect('editing-started', self.on_group_editing_started)
		group_renderer.connect('editing-canceled', self.on_group_editing_cancelled)
		group_column = Gtk.TreeViewColumn(_('Groups'), group_renderer, text=0)
		self.groups_treeview.append_column(group_column)
		groups_scroll = CompatScrolledWindow()
		groups_scroll.set_size_request(-1, 150)
		groups_scroll.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
		groups_scroll.set_child(self.groups_treeview)
		self.attach(groups_scroll, 0, 0, 2, 2)

		self.remove_group_button = Gtk.Button(label=_('Remove group'))
		self.remove_group_button.connect('clicked', self.on_delete_group_clicked)
		self.attach(self.remove_group_button, 0, 3, 1, 1)
		add_group_button = Gtk.Button(label=_('Add group'))
		add_group_button.connect('clicked', self.on_add_group_clicked)
		self.attach(add_group_button, 1, 3, 1, 1)

		# Users
		self.users_store = Gtk.ListStore(str)
		self.users_treeview = Gtk.TreeView(model=self.users_store)
		user_column = Gtk.TreeViewColumn(_('Users'), Gtk.CellRendererText(), text=0)
		self.users_treeview.append_column(user_column)
		self.users_treeview.connect('row-activated', self.on_activate_user)
		users_scroll = CompatScrolledWindow()
		users_scroll.set_size_request(-1, 150)
		users_scroll.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
		users_scroll.set_child(self.users_treeview)
		self.attach(users_scroll, 2, 0, 2, 2)

		self.remove_user_button = Gtk.Button(label=_('Remove user'))
		self.remove_user_button.connect('clicked', self.on_delete_user_clicked)
		self.attach(self.remove_user_button, 2, 3, 1, 1)
		self.users_treeview.connect('cursor-changed', self.on_user_selected)
		add_user_button = Gtk.Button(label=_('Add user'))
		add_user_button.connect('clicked', self.on_add_user_clicked)
		self.attach(add_user_button, 3, 3, 1, 1)


	def select_group(self, group: str):
		path = None
		current_iter = self.groups_store.get_iter_first()
		while path == None and current_iter != None:
			if self.groups_store.get_value(current_iter, 0) == group:
				path = self.groups_store.get_path(current_iter)
			current_iter = self.groups_store.iter_next(current_iter)
		if path == None:
			path = self.groups_store.get_path(self.groups_store.get_iter_first())
		self.groups_treeview.get_selection().select_path(path)
		self.on_group_selected(self.groups_treeview, path, self.groups_treeview.get_column(0))


	def get_selected_group(self) -> str:
		group = None
		tree_model, active_iter = self.groups_treeview.get_selection().get_selected()
		if active_iter != None:
			group = self.groups_store.get_value(active_iter, 0)
			if group == NO_GROUP:
				group = None
		return group


	def on_group_selected(self, treeview: Gtk.TreeView, path: Gtk.TreePath, column: Gtk.TreeViewColumn):
		self.error_label.set_text('')
		group = self.groups_store.get_value(self.groups_store.get_iter(path), 0)
		self.users_store.clear()
		if group == NO_GROUP:
			group = None
		count = 0
		for user in load_users(group):
			self.users_store.append([user])
			count += 1
		self.remove_user_button.set_sensitive(False)
		self.remove_group_button.set_sensitive(group != None and count == 0)


	def on_group_editing_started(self, cell: Gtk.CellRenderer, editable: Gtk.CellEditable, path: str): #Gtk.CellEditable -> Gtk.Entry
		group = self.groups_store.get_value(self.groups_store.get_iter_from_string(path), 0)
		if group == NO_GROUP and isinstance(editable, Gtk.CellEditable):
			editable.editing_done() #FIXME - Gtk-CRITICAL (but works)


	def on_group_editing_cancelled(self, cell: Gtk.CellRendererText):
		tree_model, active_iter = self.groups_treeview.get_selection().get_selected()
		group = self.groups_store.get_value(active_iter, 0)
		if group == '':
			self.groups_store.remove(active_iter)

	def on_group_edited(self, cell: Gtk.CellRendererText, path: Gtk.TreePath, new_name: str):
		self.error_label.set_text('')
		group = self.groups_store.get_value(self.groups_store.get_iter(path), 0)
		if group == '': #new group
			if self.check_group(new_name):
				self.groups_store.set_value(self.groups_store.get_iter(path), 0, new_name)
				add_group(new_name)
				self.on_group_selected(self.groups_treeview, path, self.groups_treeview.get_column(0))
			else: #abort
				self.groups_store.remove(self.groups_store.get_iter(path))
		else:
			if group != new_name and self.check_group(new_name):
				self.groups_store.set_value(self.groups_store.get_iter(path), 0, new_name)
				rename_group(group, new_name)


	def check_group(self, group: str) -> bool:
		status = check_name(group, True)
		message = [ None, _('Invalid group name: ') + _('only 32  _ , a-z, A-Z and 0-9 characters are allowed.'), _('Group name is already used.') ][status]
		if message != None:
			self.error_label.set_markup('<span foreground="red" weight="bold">' + message + '</span>')
		return True if message == None else False


	def on_add_group_clicked(self, button: Gtk.Button):
		self.error_label.set_text('')
		new_iter = self.groups_store.append([''])
		self.groups_treeview.set_cursor(self.groups_store.get_path(new_iter), self.groups_treeview.get_column(0), True)


	def on_delete_group_clicked(self, button: Gtk.Button):
		self.error_label.set_text('')
		tree_model, active_iter = self.groups_treeview.get_selection().get_selected()
		if active_iter != None:
			group = self.groups_store.get_value(active_iter, 0)
			self.confirm(_(f"Do You really want to delete {group} group?"), self.do_group_removal, (active_iter, group))


	def do_group_removal(self, data: Tuple[Gtk.TreeIter, str]):
		active_iter, group = data
		self.groups_store.remove(active_iter)
		self.select_group(NO_GROUP)
		delete_group(group)


	def on_user_selected(self, treeview: Gtk.TreeView):
		self.error_label.set_text('')
		tree_model, active_iter = self.users_treeview.get_selection().get_selected()
		if active_iter == None:
			self.remove_user_button.set_sensitive(False)
		else:
			login = self.users_store.get_value(active_iter, 0)
			self.remove_user_button.set_sensitive(not (is_logged_in(login) or login == 'root'))


	def on_activate_user(self, treeview: Gtk.TreeView, path: Gtk.TreePath, column: Gtk.TreeViewColumn):
		self.error_label.set_text('')
		login = self.users_store.get_value(self.users_store.get_iter(path), 0)
		self.app.on_edit_user(login, self.get_selected_group())


	def on_add_user_clicked(self, button: Gtk.Button):
		self.error_label.set_text('')
		self.app.on_edit_user(None, self.get_selected_group())


	def on_delete_user_clicked(self, button: Gtk.Button):
		self.error_label.set_text('')
		tree_model, active_iter = self.users_treeview.get_selection().get_selected()
		if active_iter != None:
			login = self.users_store.get_value(active_iter, 0)
			self.confirm(_(f"Do You really want to delete user {login}?"), self.do_user_removal, (active_iter, login))


	def do_user_removal(self, data: Tuple[Gtk.TreeIter, str]):
		active_iter, login = data
		self.users_store.remove(active_iter)
		delete_user(login)
		# reload and reselect group:
		# - deleting a user may result in the deletion of the group with the same name
		# - after deleting a user, the group may become empty and, therefore, deletable
		group = self.get_selected_group()
		self.groups_store.clear()
		self.groups_store.append([NO_GROUP])
		for group in load_user_groups():
			self.groups_store.append([group])
		self.select_group(group)


	def confirm(self, question: str, callback, data: str):
		self.confirm_dialog = CompatWindow(title=_('Confirm action'))
		self.confirm_dialog.set_transient_for(self.app.win)
		box = CompatBox(orientation = Gtk.Orientation.VERTICAL, spacing=10)
		box.set_margin_start(10); box.set_margin_end(10); box.set_margin_top(10); box.set_margin_bottom(10)
		self.confirm_dialog.set_child(box)
		box.append(Gtk.Label(label=question))
		buttons = CompatBox(orientation = Gtk.Orientation.HORIZONTAL, spacing=10)
		cancel_button = Gtk.Button(label=_('No'))
		cancel_button.set_hexpand(True)
		cancel_button.connect('clicked', lambda bt: self.confirm_dialog.set_visible(False))
		buttons.append(cancel_button)
		ok_button = Gtk.Button(label=_('Yes'))
		ok_button.set_hexpand(True)
		ok_button.connect('clicked', lambda bt: self.confirm_dialog.set_visible(False))
		ok_button.connect('clicked', lambda bt: callback(data))
		buttons.append(ok_button)
		self.confirm_dialog.set_default_widget(cancel_button)
		box.append(buttons)
		self.confirm_dialog.present()





class AccountsManager(Gtk.Application):
	def __init__(self):
		self.login = get_username()
		if self.login == 'root': #launched as root (sudo, pkexec, ...)
			self.login = None
		self.group = NO_GROUP
		super().__init__(application_id='org.gsettings.accounts') #flags=Gio.ApplicationFlags.DEFAULT_FLAGS
		gettext.textdomain('gsettings-accounts')
		self.connect('activate', self.on_activate)


	def on_activate(self, app):
		self.win = CompatApplicationWindow(application=app)
		self.win.set_default_icon_name('gsettings-accounts')
		header_bar = CompatHeaderBar()
		header_bar.set_title_widget(Gtk.Label(label=_('User Accounts Management')))
		about_button = CompatButton.new_from_icon_name('help-about')
		about_button.connect('clicked', self.show_about_dialog)
		header_bar.pack_start(about_button)
		self.win.set_titlebar(header_bar)
		self.win.set_resizable(False)
		if self.login == None: #this app can be used as an account manager
			self.win.set_child(ManageLists(self, self.group))
		else: #or to edit the logged in user
			self.win.set_child(EditUser(self, self.login, None))
		self.win.present()


	def show_about_dialog(self, widget):
		about_dialog = CompatAboutDialog(transient_for=self.win, modal=True)
		about_dialog.set_program_name('GSettings-Accounts')
		about_dialog.set_version('v0.4')
		about_dialog.set_authors(['Frank ENDRES <first@last.fr>'])
		about_dialog.set_translator_credits('\n- de: Frank ENDRES\n- fr: Frank ENDRES')
		about_dialog.set_website('https://gsettings.tuxfamily.org')
		about_dialog.set_license('- CeCILL (Ce[a] C[nrs] I[nria] L[ogiciel] L[ibre])\n- GPL (General Public License)')
		about_dialog.set_logo_icon_name('gsettings-accounts')
		about_dialog.set_visible(True)


	def on_edit_user(self, login: str, group: str):
		self.group = group
		self.win.set_child(EditUser(self, login, group))
		self.win.present()


	def on_end_edit_user(self):
		if self.login == None:
			self.win.set_child(ManageLists(self, self.group))
			self.win.present()
		else:
			self.quit()





# Minimalistic account library
LOGIN = '-l'; UID = '-u'; GROUP = '-g'; SHELL = '-s'; COMMENT = '-c'; MEMBERSHIP = '-G'; HOME = '-d'; PASSWORD = '-p'; LOGGED = '8'; AVATAR = ';-)' #see useradd/mod

def log_action(grp: str, cmd: str):  #TODO
	who = get_username()
	print (who + '>', '(' + grp + ')', cmd)


def get_login_defs(): #parses /etc/login.defs
	defs = { 'UID_MIN': 1000, 'UID_MAX': 9999, 'GID_MIN': 1000, 'GID_MAX': 9999, 'SYS_GID_MIN': 500, 'SYS_GID_MAX': 999, 'CHFN_RESTRICT': 'frwh' }
	keys = defs.keys()
	f = open('/etc/login.defs', 'r')
	line = f.readline()
	while line != '':
		if line.strip() != '' and not line.startswith('#'):
			key_val = line[:-1].split()
			if key_val[0] in keys:
				defs[key_val[0]] = key_val[1] if key_val[0] == 'CHFN_RESTRICT' else int(key_val[1])
		line = f.readline()
	f.close()

	return defs


def check_name(name: str, group: bool) -> int: #checks the validity of a group or username
		status = 0
		if not re.match(r'^[a-zA-Z0-9_]+$', name) or len(name) > 32:
			status = 1
		elif name in (load_groups() if group else load_users()):
			status = 2
		return status


def get_username() -> str: #returns the name of the user running this application
	return pwd.getpwuid(os.getuid()).pw_name


def is_logged_in(login: str) -> bool: #indicates whether the user is logged in
	return login in subprocess.run(['users'], check=False, capture_output=True).stdout.decode('utf-8')[:-1].split(' ')


def load_users(group: str = None) -> List[str]: #returns only non-system users
	#CHOICE: don't use pwd.getpwall()
	gid = -1
	if group != None:
		gid = grp.getgrnam(group).gr_gid
	users = []
	f = open('/etc/passwd', 'r')
	user = f.readline()
	while user != '':
		login, password, uid, user_gid, comment, home, shell = user[:-1].split(':')
		if (login == 'root' or home[:6] == '/home/') and (group == None or int(user_gid) == gid):
			users.append(login)
		user = f.readline()
	f.close()
	users.sort()
	return users


def load_shells() -> List[str]: #returns the list of the available shells
	shells = [] #CHOICE: don't read /etc/shells content
	for shell in ['bash', 'dash', 'fish', 'ksh', 'tcsh', 'zsh']: #first shell will be the default one for new users
		for prefix in ['/bin/', '/usr/bin/']:
			if os.path.exists(prefix + shell):
				shells.append(prefix + shell)
				break
	return shells


def get_emty_user() ->Dict:
	return { LOGIN: None, UID: None, GROUP: None, COMMENT: None, SHELL: None, MEMBERSHIP: None, LOGGED: False, AVATAR: None }


def get_user(login_or_uid: str) -> Dict:
	infos = None
	user = subprocess.run(['getent', 'passwd', login_or_uid], check=False, capture_output=True).stdout.decode('utf-8')[:-1] #CHOICE: don't use pwd.getpwuid() / getpwnam()
	if user != '':
		parts = user.split(':')
		name = parts[4]
		if name.find(',') != -1:
			name = name[:name.find(',')]
		infos = { LOGIN: parts[0], UID: int(parts[2]), GROUP: grp.getgrgid(int(parts[3])).gr_name, COMMENT: name, SHELL: parts[6] }
		groups = subprocess.run(['groups', infos[LOGIN]], check=False, capture_output=True).stdout.decode('utf-8')[:-1].split(' ') #CHOICE: don't use grp.getgrall()
		groups.remove(infos[GROUP])
		infos[MEMBERSHIP] = ','.join(groups)
		infos[LOGGED] = is_logged_in(infos[LOGIN])
		infos[AVATAR] = get_avatar(infos[LOGIN])
	return infos


def save_user(infos: Dict):
	params = []

	if infos[UID] == None: #new user
		cmd = 'useradd'
		login = infos[LOGIN]
		params += ['-m']
		for opt in [COMMENT, GROUP, SHELL, MEMBERSHIP]:
			if infos[opt] != '':
				params += [ opt, infos[opt] ]

	else: #existing user
		existing_infos = get_user(str(infos[UID]))
		cmd = 'usermod'
		login = existing_infos[LOGIN]
		if infos[LOGIN] != login: #if login is changed
			params += [LOGIN, infos[LOGIN], HOME, '/home/' + infos[LOGIN]] #update both login and home
			os.rename('/home/' + login, '/home/' + infos[LOGIN])
		for opt in [COMMENT, GROUP, SHELL, MEMBERSHIP]:
			if infos[opt] != existing_infos[opt]:
				params += [ opt, infos[opt] ]

	# new or existing user
	if len(params) != 0: #usermod could have no changes
		subprocess.run([cmd] + params + [login], check=False)
		log_action('user', ' '.join([cmd] + params + [login]))

	if infos[PASSWORD] != '':
		subprocess.run(['chpasswd'], input=login + ':' + infos[PASSWORD], text=True, check=False)
		log_action('user', ' '.join(['chpasswd', login]))

	if infos[AVATAR] != get_avatar(login): #new user has no existing infos
		set_avatar(login, infos[AVATAR])


def get_avatar(login: str) -> str:
	path = '/home/' + login + '/.face'
	if not os.path.exists(path):
		path = None
	return path


def set_avatar(login: str, filename: str):
	path = '/home/' + login + '/.face'
	if filename == None and os.path.exists(path):
		os.remove(path)
		log_action('avatar', 'rm ' + path)
	elif filename != None:
		shutil.copyfile(filename, path) #TODO: resize to 128x128
		log_action('avatar', 'cp ' + filename + ' ' + path)
		os.chmod(path, stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH)
		pw = pwd.getpwnam(login)
		os.chown(path, pw.pw_uid, pw.pw_gid)


def check_password(login: str, password) -> bool:
	return subprocess.run(['su', login], input=password+'\n', text=True, check=False).returncode == 0


def change_password(login: str, old_password: str, new_password: str) -> bool:
	return subprocess.run(['passwd', login], input=old_password+'\n'+new_password+'\n'+new_password+'\n', text=True, check=False).returncode == 0
	log_action('user', ' '.join(['passwd', login]))


def chfn_user(login: str, password: str, fullname: str):
	subprocess.run(['chfn', '-f', fullname, login], input=password+'\n', text=True, check=False)
	log_action('user', ' '.join(['chfn', '-f', fullname, login]))


def chsh_user(login: str, password: str, shell: str):
	subprocess.run(['chsh', '-s', shell, login], input=password+'\n', text=True, check=False)
	log_action('user', ' '.join(['chsh', '-s', shell, login]))


def delete_user(login: str):
	subprocess.run(['killall', '-s', '9', '-u', login], check=False)
	subprocess.run(['userdel', '-r', login], check=False)
	log_action('user', ' '.join(['userdel', '-r', login]))


def load_groups() -> List[str]: #CHOICE: don't use grp.getgrall()
	groups = [] #load all groups
	f = open('/etc/group', 'r')
	line = f.readline()
	while line != '':
		groups.append(line.split(':')[0])
		line = f.readline()
	f.close()

	return groups


def load_user_groups() -> List[str]: #returns only non-system user groups
	defs = get_login_defs()
	groups = ['root']
	f = open('/etc/group', 'r')
	line = f.readline()
	while line != '':
		parts = line.split(':')
		group = parts[0]
		gid = int(parts[2])
		if gid >= defs['GID_MIN'] and gid <= defs['GID_MAX'] or group == 'users':
			groups.append(group)
		line = f.readline()
	f.close()
	groups.sort()

	return groups


def load_membership_groups() -> List[str]: #returns a list of groups used to grant privileges to users
	#CHOICE: don't use SYS_GID_MIN/MAX
	#CHOICE: use a group preset
	membership_groups = [ 'adm', 'lp', 'root', 'scanner', 'sudo', 'sys', 'wheel' ] #SystemD UDev uses 'uaccess' tag to allow (audio, ...) device access to logged in users
	membership_groups += [ 'adbusers', 'vboxusers', 'wireshark' ]

	if os.path.exists('/etc/slackware-version'):
		membership_groups += [ 'audio', 'cdrom', 'disk', 'video' ] #no SystemD
		membership_groups += [ 'apache', 'dialout', 'netdev' ]
	elif os.path.exists('/etc/arch-release'): #cf https://wiki.archlinux.org/title/Users_and_groups#User_groups
		membership_groups += [ 'http', 'log', 'rfkill' ]
	elif os.path.exists('/etc/debian_version'): #cf https://wiki.debian.org/SystemGroups, https://wiki.ubuntu.com/Security/Privileges
		membership_groups += [ 'dialout', 'dip', 'netdev', 'www-data' ]

	groups = load_groups() #all groups
	for group in membership_groups:
		if group not in groups:
			membership_groups.remove(group)
	membership_groups.sort()

	return membership_groups


def add_group(group: str):
	subprocess.run(['groupadd', group], check=False)
	log_action('group', ' '.join(['groupadd', group]))


def rename_group(group: str, new_name: str):
	subprocess.run(['groupmod', '-n', new_name, group], check=False)
	log_action('group', ' '.join(['groupmod', '-n', new_name, group]))


def delete_group(group: str):
	subprocess.run(['groupdel', group], check=False)
	log_action('group', ' '.join(['groupdel', group]))





if __name__ == '__main__':
	app = AccountsManager()
	app.run([])
